
#include "include/types.h"
#include "include/riscv.h"
#include "include/param.h"
#include "include/memlayout.h"
#include "include/spinlock.h"
#include "include/proc.h"
#include "include/syscall.h"
#include "include/timer.h"
#include "include/kalloc.h"
#include "include/string.h"
#include "include/printf.h"
#include "include/times.h"
#include "include/uname.h"
#include "include/sbi.h"
#include "include/gettimeofday.h"
#include "include/nanosleep.h"
#include "include/defs.h"

extern int exec(char *path, char **argv);

uint64
sys_exec(void)
{
  char path[FAT32_MAX_PATH], *argv[MAXARG];
  int i;
  uint64 uargv, uarg;

  if(argstr(0, path, FAT32_MAX_PATH) < 0 || argaddr(1, &uargv) < 0){
    return -1;
  }
  memset(argv, 0, sizeof(argv));
  for(i=0;; i++){
    if(i >= NELEM(argv)){
      goto bad;
    }
    if(fetchaddr(uargv+sizeof(uint64)*i, (uint64*)&uarg) < 0){
      goto bad;
    }
    if(uarg == 0){
      argv[i] = 0;
      break;
    }
    argv[i] = kalloc();
    if(argv[i] == 0)
      goto bad;
    if(fetchstr(uarg, argv[i], PGSIZE) < 0)
      goto bad;
  }

  int ret = exec(path, argv);

  for(i = 0; i < NELEM(argv) && argv[i] != 0; i++)
    kfree(argv[i]);

  return ret;

 bad:
  for(i = 0; i < NELEM(argv) && argv[i] != 0; i++)
    kfree(argv[i]);
  return -1;
}

uint64
sys_exit(void)
{
  int n;
  if(argint(0, &n) < 0)
    return -1;
  exit(n);
  return 0;  // not reached
}

uint64
sys_getpid(void)
{
  return myproc()->pid;
}

uint64
sys_fork(void)
{
  return fork();
}

uint64
sys_wait(void)
{
  uint64 p;
  if(argaddr(0, &p) < 0)
    return -1;
  return wait(p);
}

uint64
sys_sbrk(void)
{
  int addr;
  int n;

  if(argint(0, &n) < 0)
    return -1;
  addr = myproc()->sz;
  if(growproc(n) < 0)
    return -1;
  return addr;
}

uint64
sys_sleep(void)
{
  int n;
  uint ticks0;

  if(argint(0, &n) < 0)
    return -1;
  acquire(&tickslock);
  ticks0 = ticks;
  while(ticks - ticks0 < n){
    if(myproc()->killed){
      release(&tickslock);
      return -1;
    }
    sleep(&ticks, &tickslock);
  }
  release(&tickslock);
  return 0;
}

uint64
sys_kill(void)
{
  int pid;

  if(argint(0, &pid) < 0)
    return -1;
  return kill(pid);
}

// return how many clock tick interrupts have occurred
// since start.
uint64
sys_uptime(void)
{
  uint xticks;

  acquire(&tickslock);
  xticks = ticks;
  release(&tickslock);
  return xticks;
}

uint64
sys_trace(void)
{
  int mask;
  if(argint(0, &mask) < 0) {
    return -1;
  }
  myproc()->tmask = mask;
  return 0;
}

uint64
sys_times(void) {
    struct tms *tms;
    if (argaddr(0, (uint64*)&tms) < 0) {
        return -1;
    }

    struct proc *p = myproc();
    tms->tms_utime = p->tms_utime;
    tms->tms_stime = p->tms_stime;
    tms->tms_cutime = p->tms_cutime;
    tms->tms_cstime = p->tms_cstime;

    return ticks;
}

uint64
sys_uname(void) {
    uint64 addr;  // 用户态传递的 struct utsname 指针
    if (argaddr(0, &addr) < 0) {
        return -1;  // 获取指针失败
    }

    struct utsname uname_kernel;  // 内核态的 utsname 结构体

    // 填充内核态的 utsname 结构体
    strncpy(uname_kernel.sysname, "xv6", sizeof(uname_kernel.sysname));
    strncpy(uname_kernel.nodename, "localhost", sizeof(uname_kernel.nodename));
    strncpy(uname_kernel.release, "1.0", sizeof(uname_kernel.release));
    strncpy(uname_kernel.version, "xv6-k210", sizeof(uname_kernel.version));
    strncpy(uname_kernel.machine, "riscv64", sizeof(uname_kernel.machine));
    strncpy(uname_kernel.domainname, "(none)", sizeof(uname_kernel.domainname));

    // 将内核态的 utsname 结构体复制到用户态
    if (copyout(myproc()->pagetable, addr, (char *)&uname_kernel, sizeof(uname_kernel)) < 0) {
      return -1;  // 复制失败
    }

    return 0;  // 成功返回 0
}

uint64
sys_clone(void)
{
  int flags;
  uint64 stack;
  uint64 ptid;
  uint64 tls;
  uint64 ctid;

  // 从用户空间获取参数
  if (argint(0, &flags) < 0 ||      // 获取 flags
      argaddr(1, &stack) < 0 ||     // 获取 stack
      argaddr(2, &ptid) < 0 ||      // 获取 ptid
      argaddr(3, &tls) < 0 ||       // 获取 tls
      argaddr(4, &ctid) < 0) {      // 获取 ctid
    return -1; // 如果获取参数失败，返回错误
  }
  // 调用 clone 函数
  return clone(flags, (void*)stack, (int*)ptid, (void*)tls, (int*)ctid);
}

uint64
sys_wait4(void)
{
  //printf("here\n");
  int pid;
  uint64 status;
  int options;

  if (argint(0, &pid) < 0 || argaddr(1, &status) < 0 || argint(2, &options) < 0)
    return -1;

  return wait4(pid, (int*)status, options);
}

uint64 
sys_execve(void) 
{
  char path[MAXPATH];
  uint64 uargv, uenvp;
  char *argv[MAXARG], *envp[MAXARG];
  int i;

  if (argstr(0, path, sizeof(path)) < 0)
      return -1;
  if (argaddr(1, &uargv) < 0)
      return -1;
  if (argaddr(2, &uenvp) < 0)
      return -1;

  // 复制 argv
  for (i = 0; i < MAXARG; i++) {
      uint64 uarg;
      if (fetchaddr(uargv + sizeof(uint64) * i, (uint64 *)&uarg) < 0)
          return -1;
      if (uarg == 0) {
          argv[i] = 0;
          break;
      }
      if (fetchstr(uarg, argv[i], PGSIZE) < 0)
          return -1;
  }

  // 复制 envp
  for (i = 0; i < MAXARG; i++) {
      uint64 uenv;
      if (fetchaddr(uenvp + sizeof(uint64) * i, (uint64 *)&uenv) < 0)
          return -1;
      if (uenv == 0) {
          envp[i] = 0;
          break;
      }
      if (fetchstr(uenv, envp[i], PGSIZE) < 0)
          return -1;
  }

  return exec(path, argv);
}

uint64
sys_getppid(void)
{
  return myproc()->parent->pid;
}

uint64
sys_sched_yield(void)
{
  struct proc *p = myproc();
  if (p == 0) {
    return -1;  // 当前进程无效，返回-1
  }
  yield();
  return 0;
}

uint64 
sys_gettimeofday(void) 
{
    uint64 ts_addr;
    struct timeval ts;

    if (argaddr(0, &ts_addr) < 0)
        return -1;

    uint64 ticks = r_time(); // 读取当前时间，单位为 ticks
    ts.sec = ticks / 100;
    ts.usec = (ticks % 100) * (1000000 / 100);

    if (copyout(myproc()->pagetable, ts_addr, (char *)&ts, sizeof(ts)) < 0){
      printf("error\n");
      return -1;
    }
    return 0;
}

uint64
sys_nanosleep(void)
{
  //printf("nanosleep here\n");
  struct timespec req, rem;
  struct timespec *reqp, *remp;

  // 从用户空间获取 req 和 rem 指针
  if (argaddr(0, (uint64*)&reqp) < 0 || argaddr(1, (uint64*)&remp) < 0)
      return -1;

  // 将用户空间的 req 结构体复制到内核空间
  if (copyin(myproc()->pagetable, (char*)&req, (uint64)reqp, sizeof(req)) < 0)
      return -1;

  // 检查纳秒字段的范围
  if (req.tv_nsec < 0 || req.tv_nsec >= 1000000000) {
      return -1;  // 无效的纳秒值
  }

  // 计算总睡眠时间（毫秒）
  uint64 total_ms = req.tv_sec * 1000 + req.tv_nsec / 1000000;
  uint64 start_ticks = r_time();

  // 调用 sleep 函数（xv6 的 sleep 以毫秒为单位）
  acquire(&myproc()->lock);
  sleep(myproc(), &myproc()->lock);
  release(&myproc()->lock);

  // 检查是否被信号中断
  if (myproc()->killed) {
    uint64 end_ticks = r_time();
    uint64 ticks = end_ticks - start_ticks;

    if (remp) {
        // 计算剩余时间
        uint64 remaining_ms = total_ms - ticks;
        rem.tv_sec = remaining_ms / 1000;
        rem.tv_nsec = (remaining_ms % 1000) * 1000000;

        // 将 rem 结构体复制回用户空间
        if (copyout(myproc()->pagetable, (uint64)remp, (char*)&rem, sizeof(rem)) < 0)
            return -1;
    }
    return -1;  // 被信号中断
  }

  return 0;  // 成功
}